/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wF1Ct.h"
using namespace tbox;
using namespace flow;

F1CtDrag::F1CtDrag(int _nDim, double _sRef, double alpha, double _beta) : F1Ct(), nDim(_nDim), sRef(_sRef), beta(_beta)
{
    // Sanity check
    if (nDim != 2 && nDim != 3)
    {
        std::stringstream err;
        err << "F1CtDrag::nDim should be equal to 2 or 3 but " << nDim << " was given!\n";
        throw std::runtime_error(err.str());
    }
    if (nDim != 3 && beta != 0)
        throw std::runtime_error("F1CtDrag::beta can be nonzero only for 3D cases!\n");
    // Set the velocity
    this->update(alpha);
}
/**
 * @brief Update the direction and its gradient
 */
void F1CtDrag::update(double alpha)
{
    if (nDim == 2)
    {
        v = Eigen::Vector3d(cos(alpha), sin(alpha), 0) / sRef;
        dv = Eigen::Vector3d(-sin(alpha), cos(alpha), 0) / sRef;
    }
    else
    {
        v = Eigen::Vector3d(cos(alpha) * cos(beta), sin(beta), sin(alpha) * cos(beta)) / sRef;
        dv = Eigen::Vector3d(-sin(alpha) * cos(beta), sin(beta), cos(alpha) * cos(beta)) / sRef;
    }
}
/**
 * @brief Return the direction
 */
Eigen::Vector3d F1CtDrag::eval() const
{
    return v;
}
/**
 * @brief Return the gradient of the direction with respect to angle of attack
 */
Eigen::Vector3d F1CtDrag::evalGrad() const
{
    return dv;
}

F1CtSide::F1CtSide(int _nDim, double _sRef, double alpha, double _beta) : F1Ct(), nDim(_nDim), sRef(_sRef), beta(_beta)
{
    // Sanity check
    if (nDim != 2 && nDim != 3)
    {
        std::stringstream err;
        err << "F1CtSide::nDim should be equal to 2 or 3 but " << nDim << " was given!\n";
        throw std::runtime_error(err.str());
    }
    if (nDim != 3 && beta != 0)
        throw std::runtime_error("F1CtSide::beta can be nonzero only for 3D cases!\n");
    // Set the velocity
    this->update(alpha);
}
/**
 * @brief Update the direction and its gradient
 */
void F1CtSide::update(double alpha)
{
    if (nDim == 2)
    {
        v = Eigen::Vector3d(0, 0, 0);
        dv = Eigen::Vector3d(0, 0, 0);
    }
    else
    {
        v = Eigen::Vector3d(-cos(alpha) * sin(beta), cos(beta), -sin(alpha) * sin(beta)) / sRef;
        dv = Eigen::Vector3d(sin(alpha) * sin(beta), cos(beta), -cos(alpha) * sin(beta)) / sRef;
    }
}
/**
 * @brief Return the direction
 */
Eigen::Vector3d F1CtSide::eval() const
{
    return v;
}
/**
 * @brief Return the gradient of the direction with respect to angle of attack
 */
Eigen::Vector3d F1CtSide::evalGrad() const
{
    return dv;
}

F1CtLift::F1CtLift(int _nDim, double _sRef, double alpha, double _beta) : F1Ct(), nDim(_nDim), sRef(_sRef), beta(_beta)
{
    // Sanity check
    if (nDim != 2 && nDim != 3)
    {
        std::stringstream err;
        err << "F1CtLift::nDim should be equal to 2 or 3 but " << nDim << " was given!\n";
        throw std::runtime_error(err.str());
    }
    if (nDim != 3 && beta != 0)
        throw std::runtime_error("F1CtLift::beta can be nonzero only for 3D cases!\n");
    // Set the velocity
    this->update(alpha);
}
/**
 * @brief Update the direction and its gradient
 */
void F1CtLift::update(double alpha)
{
    if (nDim == 2)
    {
        v = Eigen::Vector3d(-sin(alpha), cos(alpha), 0) / sRef;
        dv = Eigen::Vector3d(-cos(alpha), -sin(alpha), 0) / sRef;
    }
    else
    {
        v = Eigen::Vector3d(-sin(alpha), 0, cos(alpha)) / sRef;
        dv = Eigen::Vector3d(-cos(alpha), 0, -sin(alpha)) / sRef;
    }
}
/**
 * @brief Return the direction
 */
Eigen::Vector3d F1CtLift::eval() const
{
    return v;
}
/**
 * @brief Return the gradient of the direction with respect to angle of attack
 */
Eigen::Vector3d F1CtLift::evalGrad() const
{
    return dv;
}
