#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute adjoint solution of lifting (linear or nonlinear) potential flow around a NACA 0012
# Adrien Crovato
#
# Test the adjoint solver
#
# CAUTION
# Adjoint solver has NOT been VALDIATED.
# This test is provided to ensure that the adjoint solver runs.
# Mesh refinement may have to be performed to obtain physical results.

import numpy as np
import flow as flo
import flow.utils as floU
import flow.default as floD
import tbox.utils as tboxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 3*np.pi/180
    M_inf = 0.3
    c_ref = 1
    span = 1
    dim = 3

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'spn' : span, 'lgt' : 3, 'wdt' : 3, 'hgt' : 3, 'msLe' : 0.02, 'msTe' : 0.02, 'msF' : 1.}
    msh, gmshWriter = floD.mesh(dim, 'models/n0012_3.geo', pars, ['field', 'wing', 'symmetry', 'downstream'], wktp = 'wakeTip')
    # create mesh deformation handler
    morpher = floD.meshMorpher(msh, dim, ['wing'])
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, _ = floD.problem(msh, dim, alpha, 0., M_inf, span*c_ref, c_ref, 0., 0., 0., 'wing', tp = 'teTip')
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    solver = floD.newton(pbl)
    solver.run()
    solver.save(gmshWriter)
    tms['solver'].stop()

    # solve adjoint problem
    print(ccolors.ANSI_BLUE + 'PyGradient...' + ccolors.ANSI_RESET)
    tms['adjoint'].start()
    adjoint = floD.adjoint(solver, morpher)
    adjoint.run()
    adjoint.save(gmshWriter)
    tms['adjoint'].stop()

    # compute norm of gradient wrt to mesh
    dClX = 0
    dCdX = 0
    for n in msh.nodes:
        dClX += np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]).dot(np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]))
        dCdX += np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]).dot(np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]))
    dClX = np.sqrt(dClX)
    dCdX = np.sqrt(dCdX)

    # recover gradient wrt to AoA from mesh gradient
    drot = np.array([[-np.sin(alpha), 0, np.cos(alpha)], [0, 0, 0], [-np.cos(alpha), 0, -np.sin(alpha)]])
    dClAoA = 0
    dCdAoA = 0
    for n in bnd.nodes:
        dx = drot.dot(np.array([n.pos[0] - 1, n.pos[1], n.pos[2]]))
        dClAoA += np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]).dot(dx)
        dCdAoA += np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]).dot(dx)

    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('       M    alpha       Cl       Cd   dCl_da   dCd_da')
    print('{0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f} {5:8.4f}'.format(M_inf, alpha*180/np.pi, solver.Cl, solver.Cd, adjoint.dClAoa, adjoint.dCdAoa))

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if M_inf == 0.3 and alpha == 3*np.pi/180 and span == 1:
        tests.add(CTest('dCl/dAoA', adjoint.dClAoa, 2.5, 5e-2)) # FD 2.526 (step 1e-5)
        tests.add(CTest('dCd/dAoA', adjoint.dCdAoa, 0.146, 6e-2)) # FD 0.1460 (step 1e-5) - 0.138480 on debian11
        tests.add(CTest('dCl_dMsh', dClX, 3.778, 5e-2))
        tests.add(CTest('dCd_dMsh', dCdX, 0.192, 5e-2))
        tests.add(CTest('dCl/dAoA (msh)', dClAoA, 2.7, 5e-2))
        tests.add(CTest('dCd/dAoA (msh)', dCdAoA, 0.151, 5e-2))
    elif M_inf == 0.7 and alpha == 5*np.pi/180 and span == 2: # mesh size must be 0.01
        tests.add(CTest('dCl/dAoA', adjoint.dClAoa, 5.0, 1e-2))
        tests.add(CTest('dCd/dAoA', adjoint.dCdAoa, 0.427, 1e-2))
        tests.add(CTest('dCl_dMsh', dClX, 4.773, 1e-3))
        tests.add(CTest('dCd_dMsh', dCdX, 0.327, 1e-3))
        tests.add(CTest('dCl/dAoA (msh)', dClAoA, 5.2, 1e-2))
        tests.add(CTest('dCd/dAoA (msh)', dCdAoA, 0.441, 1e-2))
    else:
        raise Exception('Test not defined for this flow')
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
