#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Utilities for flow
# Adrien Crovato

def computeAeroCoef(_x, _y, _Cp, _alpha):
    '''Compute 2D aerodynamic coefficients
    The coefficients are computed from Cp averaged at nodes, which might leads to innacurate results
    For accurate results, use coefficients from solver or boundary objects
    '''
    import math
    i = 0
    Cy = 0
    Cx = 0
    Cm = 0 # positive nose-up (clockwise)
    while i < (len(_x)-1):
        dx = -(_x[i+1] - _x[i])
        dy = -(_y[i+1] - _y[i])
        Cy += 0.5 * dx * (_Cp[i+1] + _Cp[i])
        Cx -= 0.5 * dy * (_Cp[i+1] + _Cp[i])
        Cm -= 0.5*(_Cp[i+1]*(_x[i+1]-0.25) + _Cp[i]*(_x[i]-0.25)) * dx + 0.5*(_Cp[i+1]*_y[i+1] + _Cp[i]*_y[i]) * dy
        i = i+1
    Cl = Cy*math.cos(_alpha) - Cx*math.sin(_alpha)
    Cd = Cy*math.sin(_alpha) + Cx*math.cos(_alpha)
    return (Cl, Cd, Cm)

def convex_sort(vNodes, vData):
    '''Sort point cloud data forming a convex hull in counterclowise order, from TE
    '''
    import numpy as np
    # store map in arrays
    data = np.zeros((vNodes.size(),4))
    i = 0
    for n in vNodes:
        data[i,0] = n.pos[0]
        data[i,1] = n.pos[1]
        data[i,2] = n.pos[2]
        data[i,3] = vData[i]
        i += 1
    # sort array in counterclowise direction, from TE
    cg = np.array([np.average(data[:,0]),np.average(data[:,1])])
    angle = np.zeros((vNodes.size()))
    i = 0
    while i < len(data[:,0]):
        angle[i] = math.atan2(data[i,1]-cg[1],data[i,0]-cg[0])
        if angle[i] < 0:
            angle[i] = 2 * np.pi + angle[i]
        i+=1
    angleIdx = angle.argsort()
    data[:,0] = data[angleIdx,0]
    data[:,1] = data[angleIdx,1]
    data[:,2] = data[angleIdx,2]
    data[:,3] = data[angleIdx,3]
    max_x = np.argmax(data[:,0])
    data = np.roll(data, -(max_x+1), axis = 0)
    return data

def extract(vElems, vData):
    '''Extract data at first node of elements
    '''
    import numpy as np
    # check if input is scalar or vector
    if isinstance(vData[0], float):
        size = 1
    elif vData[0].size() == 3:
        size = 3
    else:
        raise RuntimeError('flow.utils.extract: unrecognized format for vData!')
    # store data (not efficient, but OK since only meant for small 2D cases)
    data = np.zeros((vElems.size()+1,3+size))
    i = 0
    while i < vElems.size()+1:
        data[i,0] = vElems[i%vElems.size()].nodes[0].pos[0]
        data[i,1] = vElems[i%vElems.size()].nodes[0].pos[1]
        data[i,2] = vElems[i%vElems.size()].nodes[0].pos[2]
        if size == 1:
            data[i,3] = vData[vElems[i%vElems.size()].nodes[0].row]
        else:
            for j in range(size):
                data[i,3+j] = vData[vElems[i%vElems.size()].nodes[0].row][j]
        i += 1
    return data

def writeSlices(mshName, ys, wId, n = 0):
    '''Write slice data for each ys coordinate along the span (only works with VTK)
    '''
    import numpy as np
    try:
        import vtk
    except:
        raise Exception('VTK not found!\n')
    import tbox.utils as tU
    import tboxVtk.reader as vtkR
    import tboxVtk.cutter as vtkC
    reader = vtkR.Reader()
    if n == 0:
        reader.open(mshName)
    else:
        reader.open(mshName+'_'+str(n))
    cutter = vtkC.Cutter(reader.reader.GetOutputPort())
    for i in range(0, len(ys)):
        pts, elems, vals = cutter.extract(cutter.cut(wId, [0., ys[i], 0.], [0., 1., 0.]), 2, ['Cp'])
        x_c = np.zeros((pts.shape[0],1))
        x_c[:,0] = (pts[:,0] - min(pts[:,0])) / (max(pts[:,0]) - min(pts[:,0]))
        data = np.hstack((pts, x_c))
        data = np.hstack((data, vals['Cp']))
        tU.sort(elems, data)
        data = np.vstack((data, data[0,:]))
        if n == 0:
            tU.write(data, 'slice_'+str(i)+'.dat', '%1.5e', ', ', 'x, y, z, x/c, Cp', '')
        else:
            tU.write(data, 'slice_'+str(n)+'_'+str(i)+'.dat', '%1.5e', ', ', 'x, y, z, x/c, Cp', '')
